package students.stolerman;

import java.util.*;

import edu.drexel.cs.ai.othello.*;
import edu.drexel.cs.ai.othello.GameState.*;

/**
 * Implementation of Minimax search with &alpha;-&beta; pruning. 
 * @author Ariel Stolerman
 */
public class AlphaBetaSearch
{
	/* =========================================================================
	 * Fields
	 * =========================================================================
	 */
	
	// -------------------------------------------------------------------------
	// Static fields
	// -------------------------------------------------------------------------
	
	/**
	 * The time difference threshold after which to cut off the search and
	 * return.
	 */
	public static final int DEFAULT_DEADLINE_DELTA_MS = 10;
	
	// -------------------------------------------------------------------------
	// Member fields
	// -------------------------------------------------------------------------
	
	protected Player maxPlayer;
	protected int deadlineDeltaMs;
	protected boolean printStats;
	protected HeuristicEval heuristic;
	protected int totalMoveCount;
	protected long totalMoveTime;
	protected long maxMoveTime;
	protected int maxDepthReached;
	protected int utilRetCount;
	protected int evalRetCount;
	protected int pruned;
	
	
	/* =========================================================================
	 * Constructors
	 * =========================================================================
	 */
	
	/**
	 * Default constructor for Minimax search with &alpha;-&beta; pruning
	 * object. Uses the default heuristic evaluation as defined by
	 * {@link HeuristicEval.CornerSquaresHeuristicEval()} and the deadline
	 * cutoff value defined by {@link AlphaBetaSearch#DEFAULT_DEADLINE_DELTA_MS}
	 * with statistics printout after every move set to <code>false</code>.
	 */
	public AlphaBetaSearch()
	{
		deadlineDeltaMs = DEFAULT_DEADLINE_DELTA_MS;
		heuristic = new HeuristicEval.CornerSquaresHeuristicEval();
		printStats = false;
	}
	
	/**
	 * Constructor for Minimax search with &alpha;-&beta; pruning object. Uses
	 * the given deadline delta threshold (the delta between current time and
	 * deadline in which to cut off the search and return the evaluation),
	 * heuristic evaluation function, and indicator of whether to printout move
	 * statistics after every move.
	 * @param deadlineDeltaMs
	 * @param heuristic
	 */
	public AlphaBetaSearch(int deadlineDeltaMs, HeuristicEval heuristic,
			boolean printStats)
	{
		this.deadlineDeltaMs = deadlineDeltaMs;
		this.heuristic = heuristic;
		this.printStats = printStats;
	}
	
	
	/* =========================================================================
	 * Methods
	 * =========================================================================
	 */
	
	/**
	 * Return the move selected by an &alpha;-&beta;-pruning search algorithm
	 * with the initialized heuristic. Returns an answer after the time diff
	 * between the given deadline and the time of call is less than the
	 * threshold set in construction.
	 * @param gs
	 * @param deadline
	 * @return
	 */
	public Square alphaBeta(GameState gs, Date deadline)
	{
		maxDepthReached = 0;
		utilRetCount = 0;
		evalRetCount = 0;
		pruned = 0;
		totalMoveCount++;
		long start = System.currentTimeMillis();
		maxPlayer = gs.getCurrentPlayer();
		Square move = alphaBetaRecursive(gs, gs.getCurrentPlayer(),
				Double.NEGATIVE_INFINITY,
				Double.POSITIVE_INFINITY,
				0, deadline).move;
		long time = System.currentTimeMillis() - start;
		maxMoveTime = Math.max(maxMoveTime, time);
		totalMoveTime += time;
		if (printStats)
			printStats();
		return move;
	}
	
	private MoveValue alphaBetaRecursive(GameState gs, Player player,
			double a, double b, int depth, Date deadline)
	{
		// if in terminal state or deadline cut-off, return the utility /
		// heuristic evaluation
		if (gs.getStatus() != GameStatus.PLAYING ||
				(deadline.getTime() - (new Date()).getTime()) < deadlineDeltaMs)
		{
			maxDepthReached = Math.max(maxDepthReached, depth);
			if (gs.getStatus() != GameStatus.PLAYING)
				utilRetCount++;
			else
				evalRetCount++;
			return new MoveValue(null,heuristic.eval(gs, player));
		}
		
		// Otherwise apply correct recursive call
		Player other = gs.getOpponent(player);
		Square[] validMoves =
				gs.getValidMoves(player).toArray(new Square[0]);
		MoveValue res = new MoveValue(), tmp;
		// MAX player
		if (player == maxPlayer)
		{
			for (Square move: validMoves)
			{
				tmp = alphaBetaRecursive(
						gs.applyMove(move), other, a, b, depth + 1, deadline);
				
				if (a < tmp.value)
				{
					// update alpha
					a = tmp.value;
					// update return value
					res.value = tmp.value;
					res.move = move;
				}
				// beta cut-off
				if (b <= a)
				{
					pruned++;
					break;
				}
			}
		}
		// MIN player
		else
		{
			for (Square move: validMoves)
			{
				tmp = alphaBetaRecursive(
						gs.applyMove(move), other, a, b, depth + 1, deadline);
				
				if (b > tmp.value)
				{
					// update beta
					b = tmp.value;
					// update return value
					res.value = tmp.value;
					res.move = move;
				}
				// alpha cut-off
				if (b <= a)
				{
					pruned++;
					break;
				}
			}
		}
		return res;
	}
	
	/* =========================================================================
	 * Utilities
	 * =========================================================================
	 */
	
	/**
	 * Prints out move statistics to standard output.
	 */
	public void printStats()
	{
		System.out.println();
		
		System.out.println("Move stats:");
		System.out.println("> Max depth reached: " + maxDepthReached);
		System.out.println("> Real utility values returned:   " + utilRetCount);
		System.out.println("> Heuristic evaluations returned: " + evalRetCount);
		System.out.println("> Total pruned branches: " + pruned);
		
		//System.out.println("Total moves stats:");
		//System.out.println("> Total moves: " + totalMoveCount);
		//System.out.println("> Average move time: " +
				//(totalMoveTime / totalMoveCount) + " ms");
		//System.out.println("> Max move time: " + maxMoveTime + " ms");
		
		System.out.println();
	}
	
	
	/* =========================================================================
	 * Move-value pair
	 * =========================================================================
	 */
	
	/**
	 * Class for maintaining moves and their corresponding utility value (or
	 * evaluated utility value).
	 * @author Ariel Stolerman
	 */
	public static class MoveValue
	{
		Square move;
		int value;
		
		static Comparator<MoveValue> comparator =
				new Comparator<AlphaBetaSearch.MoveValue>() {
			@Override
			public int compare(MoveValue o1, MoveValue o2) {
				return o1.value - o2.value;
			}
		};
		
		public MoveValue() {}
		
		public MoveValue(Square move, int value)
		{
			this.move = move;
			this.value = value;
		}
		
		public static Comparator<MoveValue> comparator()
		{
			return comparator;
		}
	}
}
