package students.stolerman;

import edu.drexel.cs.ai.othello.*;
import edu.drexel.cs.ai.othello.GameState.GameStatus;
import edu.drexel.cs.ai.othello.GameState.Player;

/**
 * Implementation of different heuristic evaluation functions, to estimate the
 * attractiveness of different moves.
 * @author Ariel Stolerman
 *
 */
public abstract class HeuristicEval
{
	/* =========================================================================
	 * Fields
	 * =========================================================================
	 */
	
	protected static final int NON_TERMINAL_EVAL = Integer.MAX_VALUE;
	
	
	/* =========================================================================
	 * Methods
	 * =========================================================================
	 */
	
	// Abstract
	
	/**
	 * Returns an estimation of the expected utility of the game in the given
	 * game state for the current player.
	 * @param gs
	 * @return
	 */
	public int eval(GameState gs)
	{
		return eval(gs, gs.getCurrentPlayer());
	}
	
	/**
	 * Returns an estimation of the expected utility of the game in the given
	 * game state for the given player.
	 * @param gs
	 * @return
	 */
	public abstract int eval(GameState gs, Player player);
	
	// Non-abstract
	
	/**
	 * Returns the real utility in the given game state, if it is a terminal
	 * state. Otherwise, returns NON_TERMINAL_EVAL.
	 * @param gs
	 * @param player
	 * @return
	 */
	public static int realEval(GameState gs, Player player)
	{
		if (gs.getStatus() == GameStatus.PLAYING)
			return NON_TERMINAL_EVAL;
		else
		{
			return
					gs.getScore(player) -
					gs.getScore(gs.getOpponent(player));
		}
	}
	
	
	/* =========================================================================
	 * Implementations
	 * =========================================================================
	 */
	
	/**
	 * Evaluates the expected utility based on disc count.
	 * If the game is in terminal state, the score is the real utility.
	 * @author Ariel Stolerman
	 */
	public static class DiscCountHeuristicEval extends HeuristicEval
	{
		@Override
		public int eval(GameState gs, Player player)
		{
			Player other = gs.getOpponent(player);
			return
					gs.getScore(player) -
					gs.getScore(other);
		}
	}
	
	/**
	 * Evaluates the expected utility based on valid moves count. 
	 * If the game is in terminal state, returns the real utility.
	 * @author Ariel Stolerman
	 */
	public static class ValidMovesCountHeuristicEval extends HeuristicEval
	{
		@Override
		public int eval(GameState gs, Player player)
		{
			int res = realEval(gs, player);
			if (res == NON_TERMINAL_EVAL)
			{
				Player other = gs.getOpponent(player);
				return
						gs.getValidMoves(player).size() -
						gs.getValidMoves(other).size();
			}
			else
				return res;
		}
	}
	
	/**
	 * Evaluates the expected utility based on the number of corner squares
	 * of the players. If the game is in terminal state, returns the real
	 * utility.
	 * @author Ariel Stolerman
	 *
	 */
	public static class CornerSquaresHeuristicEval extends HeuristicEval
	{
		@Override
		public int eval(GameState gs, Player player)
		{
			int res = realEval(gs, player);
			if (res != NON_TERMINAL_EVAL)
				return res;
			Player other = gs.getOpponent(player);
			return
					numCorners(gs, player) -
					numCorners(gs, other);
		}
		
		private static int[] cornerIndices = new int[]{0,7};
		
		private int numCorners(GameState gs, Player player)
		{
			int corners = 0;
			for (int i: cornerIndices)
				for (int j: cornerIndices)
					if (gs.getSquare(i, j) == player)
						corners++;
			return corners;
		}
	}
	
	/**
	 * Evaluates the expected utility based on a combination of the heuristics
	 * and their corresponding weights it is constructed with. If the game is in
	 * terminal state, returns the real utility.
	 * @author Ariel Stolerman
	 */
	public static class CumulativeHeuristicEval extends HeuristicEval
	{		
		HeuristicEval[] evals;
		double[] weights;
		
		static double DISC_COUNT_WEIGHT = 1/100;
		static double VALID_MOVES_COUNT_WEIGHT = 1;
		static double CORNERS_COUNT_WEIGHT = 10;
		
		/**
		 * Returns a default cumulative heuristic evaluation using disc count,
		 * valid moves count and corners count, with increasing weights.
		 */
		public CumulativeHeuristicEval()
		{
			evals = new HeuristicEval[]{
					new DiscCountHeuristicEval(),
					new ValidMovesCountHeuristicEval(),
					new CornerSquaresHeuristicEval()
			};
			weights = new double[]{
				DISC_COUNT_WEIGHT,
				VALID_MOVES_COUNT_WEIGHT,
				CORNERS_COUNT_WEIGHT
			};
		}
		
		public CumulativeHeuristicEval(HeuristicEval[] evals, double[] weights)
				throws IllegalArgumentException
		{
			if (evals.length != weights.length)
				throw new IllegalArgumentException(
						"Number of input heuristics must be equal to " +
						"the number of weights.");
			this.evals = evals;
			this.weights = weights;
		}
		
		@Override
		public int eval(GameState gs, Player player)
		{
			int res = realEval(gs, player);
			Player other = gs.getOpponent(player);
			if (res == NON_TERMINAL_EVAL)
			{
				res = 0;
				for (int i = 0; i < evals.length; i++)
					res +=
					evals[i].eval(gs, player) * weights[i] -
					evals[i].eval(gs, other) * weights[i];
			}
			return res;
		}
	}
}
